/**
 * Aggregation represents a set of aggregation operations on a collection.
 * @class
 */
class Aggregation {
  /**
   * Creates an instance of Aggregation.
   * @param {Array} data - The data to perform aggregation on.
   */
  constructor(data) {
    this.data = data;
    this.pipeline = [];
  }

  /**
   * Adds a $group stage to the aggregation pipeline.
   * @param {object} group - The group aggregation specification.
   * @returns {Aggregation} The Aggregation object.
   */
  group(group) {
    this.pipeline.push({ $group: group });
    return this;
  }

  /**
   * Adds a $match stage to the aggregation pipeline.
   * @param {object} match - The match aggregation specification.
   * @returns {Aggregation} The Aggregation object.
   */
  match(match) {
    this.pipeline.push({ $match: match });
    return this;
  }

  /**
   * Adds a $sort stage to the aggregation pipeline.
   * @param {object} sort - The sort aggregation specification.
   * @returns {Aggregation} The Aggregation object.
   */
  sort(sort) {
    this.pipeline.push({ $sort: sort });
    return this;
  }

  /**
   * Executes the aggregation pipeline and returns the result.
   * @returns {Array} The result of the aggregation pipeline.
   */
  execute() {
    let result = this.data;

    for (const stage of this.pipeline) {
      const stageOperator = Object.keys(stage)[0];
      const stageParams = stage[stageOperator];

      if (stageOperator === "$group") {
        result = this.groupBy(result, stageParams);
      } else if (stageOperator === "$match") {
        result = this.filter(result, stageParams);
      } else if (stageOperator === "$sort") {
        result = this.sortBy(result, stageParams);
      }
    }

    return result;
  }

  /**
   * Groups the data by the specified properties.
   * @param {Array} data - The data to group.
   * @param {object} params - The group parameters.
   * @returns {Array} The grouped data.
   */
  groupBy(data, params) {
    const groupKeys = Object.keys(params);
    const groupValues = Object.values(params);
    const groupedData = {};

    for (const item of data) {
      const groupKey = this.getGroupKey(item, groupKeys, groupValues);
      if (!groupedData[groupKey]) {
        groupedData[groupKey] = [];
      }
      groupedData[groupKey].push(item);
    }

    return Object.values(groupedData);
  }

  /**
   * Filters the data based on the specified conditions.
   * @param {Array} data - The data to filter.
   * @param {object} params - The filter parameters.
   * @returns {Array} The filtered data.
   */
  filter(data, params) {
    return data.filter((item) => this.matchesFilter(item, params));
  }

  /**
   * Sorts the data based on the specified fields.
   * @param {Array} data - The data to sort.
   * @param {object} sort - The sort parameters.
   * @returns {Array} The sorted data.
   */
  sortBy(data, sort) {
    const sortKeys = Object.keys(sort);
    const sortValues = Object.values(sort);

    return data.sort((a, b) => {
      for (let i = 0; i < sortKeys.length; i++) {
        const key = sortKeys[i];
        const value = sortValues[i];
        if (a[key] < b[key]) {
          return value;
        } else if (a[key] > b[key]) {
          return -value;
        }
      }
      return 0;
    });
  }

  /**
   * Returns the group key for an item based on the specified properties.
   * @param {object} item - The item to extract the group key from.
   * @param {Array} groupKeys - The properties to use as group keys.
   * @param {Array} groupValues - The corresponding values for the group keys.
   * @returns {string} The group key.
   */
  getGroupKey(item, groupKeys, groupValues) {
    return groupKeys.map((key) => item[key]).join("_");
  }

  /**
   * Determines if an item matches the specified filter conditions.
   * @param {object} item - The item to check against the filter.
   * @param {object} filter - The filter conditions.
   * @returns {boolean} True if the item matches the filter, false otherwise.
   */
  matchesFilter(item, filter) {
    for (const key in filter) {
      if (item[key] !== filter[key]) {
        return false;
      }
    }
    return true;
  }

  /**
   * Counts the number of documents in the collection that match the filter.
   * @param {object} filter - The filter to match the documents.
   * @returns {number} The count of matching documents.
   */
  count(filter) {
    let count = 0;
    for (const item of this.data) {
      if (this.matchesFilter(item, filter)) {
        count++;
      }
    }
    return count;
  }

  /**
   * Calculates the sum of a numeric field in the collection that match the filter.
   * @param {string} field - The field to sum.
   * @param {object} filter - The filter to match the documents.
   * @returns {number} The sum of the field values.
   */
  sum(field, filter) {
    let sum = 0;
    for (const item of this.data) {
      if (this.matchesFilter(item, filter) && typeof item[field] === "number") {
        sum += item[field];
      }
    }
    return sum;
  }

  /**
   * Calculates the average of a numeric field in the collection that match the filter.
   * @param {string} field - The field to average.
   * @param {object} filter - The filter to match the documents.
   * @returns {number} The average of the field values.
   */
  average(field, filter) {
    const matchingItems = [];
    for (const item of this.data) {
      if (this.matchesFilter(item, filter) && typeof item[field] === "number") {
        matchingItems.push(item);
      }
    }
    if (matchingItems.length === 0) {
      return 0;
    }
    const sum = matchingItems.reduce((acc, item) => acc + item[field], 0);
    return sum / matchingItems.length;
  }
}

module.exports = Aggregation;
